from pycurator.collectors.utils import saving


def test_openpyxl_save_options():
    try:
        import openpyxl
    except ImportError:
        assert 'Excel' not in saving.save_options
    else:
        assert 'Excel' in saving.save_options


def test_pyarrow_save_options():
    try:
        import pyarrow
    except ImportError:
        assert 'Parquet' not in saving.save_options \
               and 'Feather' not in saving.save_options
    else:
        assert 'Parquet' in saving.save_options \
               and 'Feather' in saving.save_options



"""
Module for validating general function inputs.
"""

from collections.abc import Callable, Collection, Iterable
from typing import Any, Optional, ParamSpec, Type, TypeVar, Union

T = TypeVar("T")
P = ParamSpec("P")
Key = TypeVar("Key")
Value = TypeVar("Value")


def sort_dict_by_keys(unsorted_dict: dict[Key, Value]) -> dict[Key, Value]:
    """Sorts a dictionary by key values."""

    return dict(map(tuple, sorted(unsorted_dict.items())))


def is_all_type(
    objects: Iterable[Any], types: Union[Type[Any], tuple[Union[Type[Any], None], ...]]
) -> bool:
    """Validate that iterable only contains objects of a given type or types.

    Parameters
    ----------
    objects : Iterable of Any
        Iterable of objects to type-check.
    types : type or tuple of types

    Returns
    -------
    bool

    Examples
    --------
    >>> is_all_type([1,2,3,4], int)
    True
    >>> is_all_type(["hello", "world", 123], (str, int))
    False
    """

    return all(isinstance(obj, types) for obj in objects)


def validate_from_arguments(
    param: str,
    func: Callable[P, T],
    validator: Callable[P, T],
    args: Optional[tuple[Any, ...]] = None,
    kwargs: Optional[dict[str, Any]] = None,
):
    """Validate arguments for wrapper functions."""

    if not (args or kwargs):
        raise ValueError("args or kwargs must be provided to extract.")

    # Checking for param in kwargs
    if kwargs:
        if not isinstance(kwargs, dict):
            raise TypeError(
                f"kwargs parameter must be of type 'dict', not '{type(kwargs)}'."
            )
        if param in kwargs:
            kwargs[param] = validator(kwargs.get(param))
    else:
        kwargs = {}

    # Checking for param in args
    if args:
        if not isinstance(args, tuple):
            raise TypeError(
                f"args parameter must be of type 'set', not '{type(args)}'."
            )
        try:
            func_params = func.__code__.co_varnames
            idx = func_params.index(param)
            if "self" in func_params:
                idx -= 1
            args = list(args)
            args[idx] = validator(args[idx])
            args = tuple(args)
        except (IndexError, ValueError):
            pass
    else:
        args = set()

    return args, kwargs


def validate_metadata_parameters(
    object_paths: Union[str, Collection[str]]
) -> Collection[str]:
    """Ensures that the metadata object paths are of the proper form.

    Parameters
    ----------
    object_paths : str or Collection of str

    Returns
    -------
    object_paths : str or Collection of str
    Raises
    ------
    TypeError
        If object paths are not all str instance.
    """

    if isinstance(object_paths, str):
        object_paths = [object_paths]
    if not all(isinstance(path, str) for path in object_paths):
        raise TypeError("All object paths must be of type str.")

    return object_paths


def validate_save_filename(filename: str) -> str:
    """Remove quotations from filename, replace spaces with underscore."""
    return filename.replace('"', "").replace("'", "").replace(" ", "_")



"""
Import term type collectors into parent package.
"""

from .dataverse import DataverseCollector
from .figshare import FigshareCollector
from .papers_with_code import PapersWithCodeCollector

try:
    from .kaggle import KaggleCollector
except (OSError, ModuleNotFoundError):
    pass



"""
Define custom typing arguments for type hinting.
"""

from typing import Any, TypeVar, Union

from pandas import DataFrame

Page = TypeVar("Page", bound="gui.ViewPage")

SearchTerm = TypeVar("SearchTerm", bound=str)
SearchType = TypeVar("SearchType", bound=str)
SearchTuple = tuple[SearchTerm, SearchType]
SearchQuery = Union[SearchType, SearchTerm, SearchTuple]

SearchResult = Union[DataFrame, None]
TermResultDict = dict[SearchTerm, SearchResult]
TypeResultDict = dict[SearchType, SearchResult]
TermTypeResultDict = dict[SearchTuple, SearchResult]
QueryResultDict = dict[SearchQuery, SearchResult]

JSONDict = dict[str, Any]

AttributeKey = TypeVar("AttributeKey", bound=str)
AttributeValue = TypeVar("AttributeValue", bound=str)
AttributeDict = dict[AttributeKey, AttributeValue]

TKVarValue = TypeVar("TKVarValue", str, int, float, bool)



import pandas as pd
import pytest

from pycurator.collectors import available_repos


def test_kaggle_availability():
    try:
        import kaggle
    except (ImportError, OSError):
        assert 'Kaggle' not in available_repos
    else:
        assert 'Kaggle' in available_repos


def test_openml_availability():
    try:
        import openml
    except ImportError:
        assert 'OpenML' not in available_repos
    else:
        assert 'OpenML' in available_repos


def test_valid_request_output(api_collector):
    url = 'https://reqres.in/api/users/2'
    r, output = api_collector.get_request_output(url=url)

    assert r.status_code == 200
    assert 'data' in output


def test_all_empty(api_collector):
    data_dict = {
        'search': pd.DataFrame(),
        'metadata': pd.DataFrame()
    }
    assert api_collector._all_empty(data_dict)


def test_not_all_empty(api_collector):
    data_dict = {
        'search': pd.DataFrame({1: [1, 2, 3], 2: [1, 3, 5]}),
        'metadata': pd.DataFrame()
    }

    assert not api_collector._all_empty(data_dict)


def test_invalid_request_output(api_collector):
    url = 'https://example.com'
    with pytest.raises(RuntimeError, match=f'Query to {url}'):
        _ = api_collector.get_request_output(url=url)


def test_search_types(type_collector):
    assert 'dataset' not in type_collector.search_type_options
    assert 'users' in type_collector.search_type_options


def test_search_type_update(type_collector):
    with pytest.raises(ValueError):
        type_collector.search_types = ('users', 'customers')



"""
Module for collecting data from Zenodo repository.
"""

import warnings
from collections.abc import Collection
from typing import Any, Optional, Iterable, NoReturn

import pandas as pd

from ..base import BaseCollector, BaseTermCollector
from pycurator._typing import SearchTerm


class ZenodoCollector(BaseTermCollector):
    """Zenodo collector for search term queries.

    Parameters
    ----------
    search_terms : list-like, optional
        Terms to search over. Can be (re)set via set_search_terms()
        or passed in directly to search functions.
    credentials : str, optional (default=None)
        JSON filepath containing credentials in form
        {repository_name}: {key}.
    """

    def __init__(
            self,
            search_terms: Optional[Collection[SearchTerm]] = None,
            credentials: Optional[bool] = None,
    ) -> None:
        super().__init__(
            repository_name="zenodo", search_terms=search_terms, credentials=credentials
        )
        self.base_url = "https://zenodo.org/api/records"

    @staticmethod
    def accepts_user_credentials() -> bool:
        return True

    @BaseTermCollector.validate_search_term
    @BaseCollector.track_indeterminate_progress
    def get_individual_search_output(self, search_term: SearchTerm) -> pd.DataFrame:
        """Returns information about all records from Zenodo.

        Parameters
        ----------
        search_term : str

        Returns
        -------
        search_df : pandas.DataFrame

        Raises
        ------
        TypeError
            Incorrect search_term type.

        Warns
        -----
        RuntimeWarning
            Unsuccessful query response from the API.
        """

        search_year = 2022
        search_df = pd.DataFrame()
        start_date = f"{search_year}-01-01"
        end_date = f"{search_year}-12-31"

        search_params = {
            "q": f"{search_term} AND created:[{start_date} TO {end_date}]",
            "page": 1,
            "size": 1000,
        }

        response, output = self.get_request_output_and_update_query_ref(
            url=self.base_url,
            params=search_params,
            search_term=search_term,
            year=search_year,
            page=search_params["page"],
        )

        # Handle any potential errors
        if response.status_code != 200:
            warnings.warn(
                f"{response.status_code}: Returning without results.", RuntimeWarning
            )
            self.continue_running = False
            self.terminate()

        while output.get("hits").get("total"):
            while response.status_code == 200 and output.get("hits").get("hits"):
                output = output["hits"]["hits"]

                output_df = pd.DataFrame(output)
                output_df["page"] = search_params["page"]

                search_df = pd.concat([search_df, output_df]).reset_index(drop=True)

                search_params["page"] += 1
                response, output = self.get_request_output_and_update_query_ref(
                    url=self.base_url,
                    params=search_params,
                    search_term=search_term,
                    year=search_year,
                    page=search_params["page"],
                )

            search_year -= 1
            start_date = f"{search_year}-01-01"
            end_date = f"{search_year}-12-31"

            search_params[
                "q"
            ] = f"{search_term} AND created:[{start_date} TO {end_date}]"
            search_params["page"] = 1

            response, output = self.get_request_output_and_update_query_ref(
                url=self.base_url,
                params=search_params,
                search_term=search_term,
                year=search_year,
                page=search_params["page"],
            )

        self.num_queries = False

        return search_df

    def get_query_metadata(self, object_paths: Iterable[Any]) -> NoReturn:
        raise NotImplementedError("Zenodo does not provide object metadata.")



import pytest

from pycurator.collectors import base


@pytest.fixture
def metadata_params():
    return ['100', '501', '200', '300']


@pytest.fixture
def sample_args(metadata_params):
    return (metadata_params, 'users')


@pytest.fixture
def sample_kwargs(metadata_params):
    return {
        'object_paths': metadata_params,
        'search_type': 'users'
    }


@pytest.fixture
def api_collector():
    return base.BaseAPICollector('reqres')


@pytest.fixture
def type_collector():
    class TestTypeCollector(base.BaseTypeCollector):
        def __init__(self):
            super().__init__(repository_name='base_term_type')
            self.base_url = 'https://reqres.in/api'

        def get_query_metadata(self, object_paths, search_type):
            pass

        def get_all_metadata(self, search_dict):
            pass

        def accepts_user_credentials(self):
            return False

        @classmethod
        @property
        def search_type_options(cls):
            return ('users',)

        def get_individual_search_output(
                self,
                search_type
        ):
            search_url = f'{self.base_url}/{search_type}'
            search_params = {'page': 1}

            return self.get_request_output(
                url=search_url,
                params=search_params
            )

    return TestTypeCollector()



"""
Module for creating model for MVC format.
"""

import threading
from typing import Any, Type, ParamSpec

from pycurator.collectors import base as collector_base

P = ParamSpec("P")


class ThreadedRun(threading.Thread):
    """Wrapper for concrete Collector object to allow threading.

    Parameters
    ----------
    collector : Subclass of BaseCollector
    **kwargs : dict, optional
        Additional parameters to pass to the run function of the given
        collector object. Current integrated examples include the
        save_type of the output file and save_dir for storing the
        output file.

    See Also
    --------
    run :
        Start-to-finish pipeline for running Collectors. See specific
        repository Collector classes for more concrete details.
    """

    def __init__(self, collector: collector_base.BaseCollector, **kwargs: Any) -> None:
        self.collector = collector
        super().__init__(target=self.collector.run, **kwargs)


class CollectorModel:
    """Model for the PyCurator UI. Extends the Collector class.

    Parameters
    ----------
    collector_class : Inherited BaseCollector class
    collector_name : str

    Attributes
    ----------
    collector_class : Inherited BaseCollector class
    collector_name : str
    collector : Inherited BaseCollector instance
    run_thread : threading.Thread
        Separate thread for data collection so that UI functionality
        is not interrupted.
    """

    def __init__(
            self, collector_class: Type[collector_base.BaseCollector], collector_name: str
    ) -> None:
        self.collector_class = collector_class
        self.collector_name = collector_name
        self.collector = None
        self.run_thread = None

        self.requirements = {
            "search_terms": issubclass(
                self.collector_class, collector_base.TermQueryMixin
            ),
            "search_types": issubclass(
                self.collector_class, collector_base.TypeQueryMixin
            ),
        }

    def initialize_collector(self, **param_val_kwargs: P.kwargs) -> None:
        """Instantiate Collector object from class and provided kwargs.

        Parameters
        ----------
        **param_val_kwargs : dict, optional
            Parameters for Collector initialization. See individual
            Collector classes for specifics.

        See Also
        --------
        pycurator.collectors : Classes for repository date collection.
        """

        self.collector = self.collector_class(**param_val_kwargs)

    def initialize_thread(self, thread_kwargs: P.kwargs) -> None:
        """Instantiate Collector Thread with runtime arguments.

        Parameters
        ----------
        thread_kwargs : dict, optional
            Parameters to provide to run function of collector.

        See Also
        -------
        pycurator.gui.base.ThreadedRun :
            Collector wrapper allowing for seamless UI performance by
            shifting collector actions to separate thread.
        """

        self.run_thread = ThreadedRun(collector=self.collector, kwargs=thread_kwargs)



import pytest

from pycurator.collectors.utils import validating


def test_validate_save_filename():
    assert validating.validate_save_filename(
        '\'output\' "file".txt'
    ) == 'output_file.txt'


def test_validate_metadata_parameters_single(metadata_params):
    assert validating.validate_metadata_parameters(
        metadata_params[0]
    ) == [metadata_params[0]]


def test_validate_metadata_parameters_multiple(metadata_params):
    assert validating.validate_metadata_parameters(
        metadata_params
    ) == metadata_params


def test_validate_metadata_parameters_error(metadata_params):
    with pytest.raises(TypeError):
        validating.validate_metadata_parameters(list(map(int, metadata_params)))


def test_validate_from_arguments_no_params(type_collector):
    with pytest.raises(ValueError):
        validating.validate_from_arguments(
            validator=type_collector._validate,
            param='search_type',
            func=type_collector.get_query_metadata
        )


def test_validate_from_arguments_args(sample_args, type_collector):
    assert validating.validate_from_arguments(
        validator=type_collector._validate,
        param='search_type',
        func=type_collector.get_query_metadata,
        args=sample_args,
    ) == (sample_args, dict())


def test_validate_from_arguments_kwargs(sample_kwargs, type_collector):
    assert validating.validate_from_arguments(
        validator=type_collector._validate,
        param='search_type',
        func=type_collector.get_query_metadata,
        kwargs=sample_kwargs,
    ) == (set(), sample_kwargs)


def test_validate_from_arguments_args_kwargs(
        sample_args,
        sample_kwargs,
        type_collector
):
    assert validating.validate_from_arguments(
        validator=type_collector._validate,
        param='search_type',
        func=type_collector.get_query_metadata,
        args=sample_args,
        kwargs=sample_kwargs
    ) == (sample_args, sample_kwargs)



"""
Module for creating base classes of UI.
"""

from __future__ import annotations

import tkinter as tk

from . import controller, view


class PyCuratorUI(tk.Tk):
    """UI for PyCurator Application.

    The UI is structured in a Model-View-Controller format. The View
    is separated into a Landing Page with license information, a
    Selection Page with repository options and related runtime
    parameters, and a Run Page providing real-time updates on the status
    of the data collection task.

    See Also
    --------
    pycurator.gui.MVC :
        Module containing the Model, View, and Controller classes.
    """

    def __init__(self) -> None:
        super().__init__()

        self.title("PyCurator")

        curator_view = view.view.CuratorView(self)
        curator_controller = controller.CuratorController(view_=curator_view)

        self.bind("<Return>", curator_controller.request_next_page)

        curator_view.set_controller(curator_controller)
        curator_view.show()



"""
Module for collecting data from OpenML repository.
"""

import ast
from collections.abc import Collection
from typing import Optional, Union

import openml
import pandas as pd

from ..base import BaseTypeCollector
from ..utils import validating
from ..._typing import SearchType, TypeResultDict


class OpenMLCollector(BaseTypeCollector):
    """OpenML collector for search type queries.

    Parameters
    ----------
    search_types : list-like, optional
        Types to search over. Can be (re)set via set_search_types() or
        passed in directly to search functions.
    credentials : str, optional (default=None)
        OpenML API key.
    """

    def __init__(
        self,
        search_types: Optional[Collection[SearchType]] = None,
        credentials: Optional[str] = None,
    ) -> None:

        super().__init__(repository_name="openml", search_types=search_types)

        self.base_url = None

        if not openml.config.apikey:
            openml.config.apikey = credentials

    @staticmethod
    def accepts_user_credentials() -> bool:
        return True

    @classmethod
    @property
    def search_type_options(cls) -> tuple[SearchType, ...]:
        return ("datasets", "runs", "tasks", "evaluations")

    @staticmethod
    def _get_evaluations_search_output() -> pd.DataFrame:
        """Retrieve information on OpenML Evaluations."""
        evaluations_measures = openml.evaluations.list_evaluation_measures()
        evaluations_df = pd.DataFrame()

        # Get evaluation data for each available measure
        for measure in evaluations_measures:
            evaluations_dict = dict(openml.evaluations.list_evaluations(measure))

            # Convert string list array_data to list
            if evaluations_dict.get("array_data"):
                try:
                    evaluations_dict["array_data"] = ast.literal_eval(
                        evaluations_dict["array_data"]
                    )
                # Occurs if 'array_data' is already a list
                except ValueError:
                    pass

            measure_df = pd.DataFrame(
                map(lambda evaluation: evaluation.__dict__, evaluations_dict.values())
            )

            evaluations_df = pd.concat([evaluations_df, measure_df]).reset_index(
                drop=True
            )

        return evaluations_df

    @BaseTypeCollector.validate_search_type
    def get_individual_search_output(self, search_type: SearchType) -> pd.DataFrame:
        """Queries OpenML API for the specified search type.

        Parameters
        ----------
        search_type : {'datasets', 'runs', 'tasks', 'evaluations'}

        Returns
        -------
        search_df : pandas.DataFrame

        Raises
        ------
        ValueError
            Incorrect search_type provided.
        """

        self.status_queue.put(f"Querying {search_type}...")

        if search_type == "evaluations":
            return self._get_evaluations_search_output()

        # Use query type to get necessary openml api functions
        base_command = getattr(openml, search_type)
        list_queries = getattr(base_command, f"list_{search_type}")

        index = 0
        size = 10000
        search_df = pd.DataFrame()

        self._update_query_ref(page=index)
        search_results = list_queries(offset=(index * size), size=size)

        while search_results:
            output_df = pd.DataFrame(search_results).transpose()
            output_df["page"] = index + 1
            search_df = pd.concat([search_df, output_df]).reset_index(drop=True)

            index += 1

            self._update_query_ref(page=index)
            search_results = list_queries(offset=(index * size), size=size)

        self.status_queue.put(f"{search_type} search complete.")

        return search_df

    @BaseTypeCollector.validate_search_type
    def get_query_metadata(
        self, object_paths: Union[str, Collection[str]], search_type: SearchType
    ) -> pd.DataFrame:
        """Retrieves the metadata for the object_paths objects.

        Parameters
        ----------
        object_paths : str or list-like
        search_type : {'datasets', 'runs', 'tasks', 'evaluations'}

        Returns
        -------
        metadata_df : pandas.DataFrame

        Raises
        ------
        ValueError
            Invalid search_type provided.
        """

        object_paths = validating.validate_metadata_parameters(object_paths)

        base_command = getattr(openml, search_type)
        get_query = getattr(base_command, f"get_{search_type[:-1:]}")

        queries = []
        error_queries = []
        for object_path in object_paths:
            self._update_query_ref(object_name=object_path)
            try:
                query = get_query(object_path).__dict__
                queries.append(query)
            except openml.exceptions.OpenMLServerException:
                error_queries.append(object_path)

        metadata_df = pd.DataFrame(queries)

        self.status_queue.put(f" {search_type} metadata query complete.")

        return metadata_df

    def get_all_metadata(self, search_dict: TypeResultDict) -> TypeResultDict:
        """Retrieves metadata for records contained in input DataFrames.

        Parameters
        ----------
        search_dict : dict
            Dictionary of DataFrames from get_all_search_outputs.

        Returns
        -------
        metadata_dict : dict
        """

        metadata_dict = {}

        for query, query_df in search_dict.items():
            self.status_queue.put(f"Querying {query} metadata...")
            if query == "datasets":
                id_name = "did"
            elif query == "runs":
                id_name = "run_id"
            elif query == "tasks":
                id_name = "tid"
            else:
                raise ValueError(f"Query '{query}' is not a valid search_type.")

            object_paths = query_df[id_name].values

            metadata_dict[query] = self.get_query_metadata(
                object_paths=object_paths, search_type=query
            )

        self.status_queue.put("Metadata query complete.")

        return metadata_dict



"""
Module for extending functionality for saving results.
"""

import os
from typing import Union

import pandas as pd

from . import validating

save_options = {"CSV": ".csv", "JSON": ".json", "Pickle": ".pkl"}
try:
    import openpyxl

    save_options["Excel"] = ".xlsx"
except ImportError:
    pass
try:
    import pyarrow

    save_options["Parquet"] = ".parquet"
    save_options["Feather"] = ".feather"
except ImportError:
    pass

save_options = validating.sort_dict_by_keys(save_options)


def save_results(
    results: dict, data_dir: Union[str, os.PathLike[str]], output_format: str
) -> None:
    """Export DataFrame objects to specified directory.

    Parameters
    ----------
    results : dict
        Dictionary of pandas.DataFrame objects to be saved.
    data_dir : str or PathLike str
        Path for parent directory to save files to.
    output_format : str
        Format for saved results. Acceptable options are seen in
        save_options.

    Raises
    ------
    TypeError
        "results" not of type dict or "data_dir" not of type str.
    ValueError
        Incorrect output_format provided.

    See Also
    --------
    os.PathLike : ABC for objects representing a file system path.
    save_options : Output format to extension dict.
    """

    if not isinstance(results, dict):
        raise TypeError(f"results must be of type dict, not '{type(results)}'.")
    if not isinstance(data_dir, str):
        raise TypeError(f"data_dir must of type str, not '{type(data_dir)}'.")

    try:
        extension = save_options[output_format]
    except KeyError as invalid_format:
        raise ValueError(
            f"output_format must be one of {list(save_options.keys())}, "
            f"not '{output_format}'."
        ) from invalid_format

    if not os.path.isdir(data_dir):
        os.makedirs(data_dir)

    for query, query_df in results.items():
        if isinstance(query, str):
            output_filename = f"{query}{extension}"
        else:
            search_term, search_type = query
            output_filename = f"{search_term}_{search_type}{extension}"

        output_filename = validating.validate_save_filename(output_filename)

        save_dataframe(
            results=query_df,
            filepath=os.path.join(data_dir, output_filename),
            output_format=output_format,
        )


def save_dataframe(
    results: pd.DataFrame, filepath: Union[str, os.PathLike[str]], output_format: str
) -> None:
    """Saves the specified results to the file provided.

    Parameters
    ----------
    results : pandas.DataFrame
        If DataFrame, results will be stored in a csv format.
    filepath : str or PathLike str
        Path to save file. Take note of output type as specified
        above, as appending the incorrect filetype may result in the
        file being unreadable.
    output_format : str
        Format for saved results. Acceptable options are seen in
        save_options.

    Raises
    ------
    ValueError
        If a non-dataframe object results or an incorrect output_format
        is provided.

    See Also


    Examples
    --------
    >>> data = {"id": [123, 432, 576, 194], "value": [0, 1, 1, 0]}
    >>> save_dataframe(
    ...     results=data,
    ...     filepath='data/id_vals.json',
    ...     output_format='JSON'
    ... )
    Traceback (most recent call last):
        ...
    ValueError: 'results' must be of type pandas.DataFrame, not 'dict'.

    >>> df1 = pd.DataFrame(data)
    >>> save_dataframe(
    ...     results=df1,
    ...     filepath='data/id_vals.json',
    ...     output_format='JSON'
    ... )

    >>> df2 = df1
    >>> df2.loc[:, 'Name'] = ('Jeff', 'Scott', 'Will', 'Dan')
    >>> save_dataframe(
    ...     results=df2,
    ...     filepath='data/updated_id_vals.csv',
    ...     output_format='CSV'
    ... )

    >>> save_dataframe(
    ...     results=df2,
    ...     filepath='data/full_data.txt',
    ...     output_format='TXT'
    ... )
    Traceback (most recent call last):
        ...
    ValueError: 'TXT' is not supported. 'output_format' must be one of ['CSV', 'JSON', 'Pickle'].
    """

    if not isinstance(results, pd.DataFrame):
        raise ValueError(
            f"'results' must be of type pandas.DataFrame, not" f" '{type(results)}'."
        )

    try:
        getattr(results, f"to_{output_format.lower()}")(filepath)
    except AttributeError as invalid_format:
        raise ValueError(
            f"'{output_format}' is not supported. 'output_format'"
            f" must be one of {list(save_options.values())}."
        ) from invalid_format



"""
Module for collecting data from Papers With Code repository.
"""

from collections.abc import Collection
from typing import Any, Optional, Union

import pandas as pd

from ..base import BaseCollector, BaseTermTypeCollector
from ..utils.validating import validate_metadata_parameters
from pycurator._typing import (
    SearchTerm,
    SearchType,
    SearchQuery,
    TermResultDict,
    TermTypeResultDict,
    TypeResultDict,
)


class PapersWithCodeCollector(BaseTermTypeCollector):
    """PapersWithCode collector for search term and type queries.

    Parameters
    ----------
    search_terms : list-like, optional (default=None)
        Terms to search over. Can be (re)set via set_search_terms()
        or passed in directly to search functions to override set
        parameter.
    search_types : list-like, optional
        Types to search over. Can be (re)set via set_search_types()
        or passed in directly to search functions.
    credentials : str, optional (default=None)
        JSON filepath containing credentials in form
        {repository_name}: {key}.
    """

    def __init__(
            self,
            search_terms: Optional[Collection[SearchTerm]] = None,
            search_types: Optional[Collection[SearchType]] = None,
            credentials: Optional[str] = None,
    ) -> None:
        super().__init__(
            repository_name="paperswithcode",
            search_terms=search_terms,
            search_types=search_types,
            credentials=credentials,
        )
        self.base_url = "https://paperswithcode.com/api/v1"

    @staticmethod
    def accepts_user_credentials() -> bool:
        return True

    @classmethod
    @property
    def search_type_options(cls) -> tuple[SearchType, ...]:
        return ("conferences", "datasets", "evaluations", "papers", "tasks")

    @BaseCollector.track_indeterminate_progress
    def _conduct_search_over_pages(
            self,
            search_url: str,
            search_params: Optional[Any] = None,
            print_progress: bool = False,
    ) -> Union[pd.DataFrame, None]:
        """Query paginated results from the Papers With Code API.

        Parameters
        ----------
        search_url : str
        search_params : dict
            Contains parameters to pass to requests.get({params}). Most
            common include search term 'q', and page index 'page'. For
            full details, see the Notes.
        print_progress : bool, optional (default=False)
            If True, updates on query page progress is sent to object
            queue to be displayed in UI window.

        Returns
        -------
        search_df : pandas.DataFrame or None

        Notes
        -----
        For querying base objects, the following parameters are most common:
        q : str, optional
            Term to query for.
        page : int, optional
            Number for page to request.
        items_per_page : int, optional
            Number of results to return.
        ordering : str, optional
            Which field to order results by.

        For more detailed parameters, visit
        https://paperswithcode.com/api/v1/docs/
        """

        search_df = pd.DataFrame()

        if print_progress:
            self._update_query_ref(page=search_params["page"])
        _, output = self.get_request_output(url=search_url, params=search_params)

        while output.get("results"):
            output = output["results"]

            output_df = pd.DataFrame(output)
            output_df["page"] = search_params["page"]

            search_df = pd.concat([search_df, output_df]).reset_index(drop=True)

            search_params["page"] += 1

            if print_progress:
                self._update_query_ref(page=search_params["page"])

            response, output = self.get_request_output(
                url=search_url, params=search_params
            )

            # Ensure we've received results if they exist
            # 200: OK, 404: page not found (no more results)
            while response.status_code not in [200, 404]:
                self.status_queue.put(
                    f'Search error "{response.status_code}" on '
                    f'page {search_params["page"]}'
                )
                search_params["page"] += 1

                if print_progress:
                    self._update_query_ref(page=search_params["page"])

                response, output = self.get_request_output(
                    url=search_url, params=search_params
                )

        if not search_df.empty:
            return search_df

    @BaseTermTypeCollector.validate_term_and_type
    def get_individual_search_output(
            self, search_term: SearchTerm, search_type: SearchType
    ) -> pd.DataFrame:
        """Queries Papers With Code API for the specified search term and type.

        Parameters
        ----------
        search_term : str
        search_type : {
            'conferences', 'datasets', 'evaluations', 'papers', 'tasks'
        }

        Returns
        -------
        pandas.DataFrame

        Raises
        ------
        TypeError
            Incorrect search_term type.
        ValueError
            Invalid search_type provided.
        """

        search_url = f"{self.base_url}/{search_type}"

        search_params = {"q": search_term, "page": 1, "items_per_page": 500}

        return self._conduct_search_over_pages(
            search_url=search_url, search_params=search_params, print_progress=True
        )

    @BaseTermTypeCollector.validate_search_type
    def _get_metadata_types(self, search_type: SearchType) -> Collection[SearchType]:
        if search_type == "conferences":
            metadata_types = ["proceedings"]
        elif search_type == "datasets":
            metadata_types = ["evaluations"]
        elif search_type == "evaluations":
            metadata_types = ["metrics", "results"]
        elif search_type == "papers":
            metadata_types = ["datasets", "methods", "repositories", "results", "tasks"]
        elif search_type == "tasks":
            metadata_types = ["children", "evaluations", "papers", "parents"]

        return metadata_types

    def get_query_metadata(
            self, object_paths: Collection[str], **kwargs: Any
    ) -> TypeResultDict:
        """Retrieves metadata for the object_paths objects.

        Parameters
        ----------
        object_paths : str or collection of str
        **kwargs : dict, optional
            Must include search_type : {
                'conferences', 'datasets', 'evaluations', 'papers', 'tasks'
            }
            Search_type in kwargs to match signature of base method.

        Returns
        -------
        metadata_dict : dict
            Results are stored in the format
            metadata_dict[metadata_type] = DataFrame
        """

        try:
            search_type = kwargs["search_type"]
        except KeyError as no_search_type:
            raise AttributeError(
                "search_type must be passed to get_query_metadata."
            ) from no_search_type
        object_paths = validate_metadata_parameters(object_paths)

        metadata_types = self._get_metadata_types(search_type)
        metadata_dict = {}

        for metadata_type in metadata_types:
            search_df = pd.DataFrame()
            self.status_queue.put(f"Querying {metadata_type}.")

            for object_path in self.track_determinate_progress(object_paths):
                search_url = (
                    f"{self.base_url}/{search_type}/{object_path}/" f"{metadata_type}"
                )
                search_params = {"page": 1}

                # Conduct the search and add supplementary info to DataFrame
                object_df = self._conduct_search_over_pages(
                    search_url=search_url, search_params=search_params
                )

                if object_df is not None:
                    object_df["id"] = object_path
                    object_df["page"] = search_params["page"]

                search_df = pd.concat([search_df, object_df]).reset_index(drop=True)

            if not search_df.empty:
                metadata_dict[metadata_type] = search_df

        return metadata_dict

    def get_all_metadata(
            self, search_dict: TermTypeResultDict
    ) -> dict[SearchQuery, TermResultDict]:
        """Retrieves metadata for records contained in input DataFrames.

        Parameters
        ----------
        search_dict : dict
            Dictionary of DataFrames from get_all_search_outputs.

        Returns
        -------
        metadata_dict : dict
        """

        metadata_dict = {}

        for query, query_df in search_dict.items():
            if query_df is not None:
                search_term, search_type = query
                self.status_queue.put(f"Querying {search_term} {search_type} metadata.")

                object_paths = query_df.id.values

                metadata_dict[query] = self.get_query_metadata(
                    object_paths=object_paths, search_type=search_type
                )

                self.status_queue.put(
                    f"{search_term} {search_type} metadata query complete."
                )

        return metadata_dict

    def merge_search_and_metadata_dicts(self, search_dict, metadata_dict, **kwargs):
        """Merges together search and metadata DataFrames by 'on' value.

        Parameters
        ----------
        search_dict : dict
            Dictionary of search output results.
        metadata_dict : dict
            Dictionary of metadata results.
        **kwargs : dict, optional
            Additional keyword arguments to pass to pandas.merge.

        Returns
        -------
        merged_dict : dict

        See Also
        --------
        pandas.merge
        """

        merged_dict = {}

        for query_key, type_df_dict in metadata_dict.items():
            search_term, search_type = query_key
            search_df = search_dict[query_key]

            for metadata_type, metadata_df in type_df_dict.items():
                _search_type = f"{search_type}_{metadata_type}"
                df_all = pd.merge(
                    left=search_df,
                    right=metadata_df,
                    how="outer",
                    suffixes=("_search", "_metadata"),
                    **kwargs,
                )

                merged_dict[(search_term, _search_type)] = df_all

        return merged_dict



"""
Module for creating UI Landing Page.
"""

import tkinter as tk
from textwrap import dedent
from tkinter import ttk
from typing import NoReturn

from . import page


class LandingPage(page.ViewPage):
    """Landing Page of the PyCurator View component.

    See Also
    --------
    pycurator.gui.base.ViewPage
    """

    @page.ViewPage.no_overwrite
    def show(self) -> None:
        """Arrange and display Landing Page of PyCurator UI."""
        label = ttk.Label(self, text="PyCurator", font="helvetica 16 bold")

        message = tk.StringVar(
            value=dedent(
                """
                Making repository research curation as easy as py.

                This software is Copyright © 2021-2022
                The Regents of the University of California.
                All Rights Reserved.
                For full licensing information, reference the LICENSE present
                in the PyCurator repository.

                The initial development of this program was funded by the
                Librarians Association of the University of California (LAUC)
                and UC San Diego Library Research Data Curation Program (RDCP).
                """
            )
        )

        message_box = tk.Message(self, textvariable=message)

        self.next_page_button = ttk.Button(
            self, text="Continue", command=lambda: self.controller.show(self.next_page)
        )

        label.grid(row=0, pady=7)
        message_box.grid(row=1, padx=10)
        self.next_page_button.grid(row=2, pady=5)
        self.grid(row=0, column=0, sticky="nsew")
        self.tkraise()

    def reset_frame(self) -> NoReturn:
        pass



"""
Module for collecting data from Kaggle repository.
"""

import json
import os
from collections.abc import Collection
from typing import Optional, Union

import pandas as pd
from kaggle import KaggleApi
from kaggle.rest import ApiException

from ..base import BaseCollector, BaseTermTypeCollector
from ..utils.validating import validate_metadata_parameters
from ..._typing import JSONDict, SearchTerm, SearchType, TermTypeResultDict


class KaggleCollector(BaseTermTypeCollector):
    """Kaggle collector for search term and type queries.

    Parameters
    ----------
    search_terms : list-like, optional (default=None)
        Terms to search over. Can be (re)set via set_search_terms() or
        passed in directly to search functions.
    search_types : list-like, optional (default=None)
        Data types to search over. Can be (re)set via set_search_types()
        or passed in directly to search functions to override set
        parameter.

    Notes
    -----
    For validating Kaggle requests, read the official documentation on
    authentication at https://www.kaggle.com/docs/api.
    """

    def __init__(
        self,
        search_terms: Optional[Collection[SearchTerm]] = None,
        search_types: Optional[Collection[SearchType]] = None,
    ) -> None:

        super().__init__(
            repository_name="kaggle",
            search_terms=search_terms,
            search_types=search_types,
        )

        self.api = KaggleApi()
        self.api.authenticate()
        self.merge_on = "id"

    @staticmethod
    def accepts_user_credentials() -> bool:
        return False

    @classmethod
    @property
    def search_type_options(cls) -> tuple[SearchType, ...]:
        return ("datasets", "kernels")

    @BaseCollector.track_indeterminate_progress
    @BaseTermTypeCollector.validate_term_and_type
    def get_individual_search_output(
        self,
        search_term: SearchTerm,
        search_type: SearchType,
    ) -> pd.DataFrame:
        """Queries the Kaggle API for the specified search term and type.

        Parameters
        ----------
        search_term : str
        search_type : {'datasets', 'kernels'}

        Returns
        -------
        search_df : pandas.DataFrame

        Raises
        ------
        TypeError
            Incorrect search_term type.
        ValueError
            Invalid search_type provided.
        """

        # Use search type to get relevant API function
        list_queries = getattr(self.api, f"{search_type}_list")

        page_idx = 1
        search_df = pd.DataFrame()

        self._update_query_ref(page="page_idx")
        output = list_queries(search=search_term, page=page_idx)

        while output:
            if not self.continue_running:
                self.terminate()

            if search_type == "kernels":
                output = [vars(result) for result in output]

            output_df = pd.DataFrame(output)
            output_df["page"] = page_idx

            search_df = pd.concat([search_df, output_df]).reset_index(drop=True)

            page_idx += 1
            self._update_query_ref(page=page_idx)
            output = list_queries(search=search_term, page=page_idx)

        if not search_df.empty:
            search_df = search_df.rename(columns={"id": "datasetId", "ref": "id"})

            if search_type == "datasets":
                search_df = search_df.drop(columns={"viewCount", "voteCount"})

            search_df = search_df.convert_dtypes()

        return search_df

    def _retrieve_object_json(
        self, object_path: str, data_path: Optional[str] = f"data{os.sep}"
    ) -> Union[JSONDict, None]:
        """Queries the Kaggle API for metadata JSON file.

        Parameters
        ----------
        object_path : str
        data_path : str, optional (default='data/')
            Location to save metadata to.

        Returns
        -------
        json_data : dict or None

        Raises
        ------
        kaggle.rest.ApiException
            A query was made that was unable to be fulfilled.

        See Also
        --------
        kaggle
        """

        if not self.continue_running:
            self.terminate()

        try:
            self.api.dataset_metadata(dataset=object_path, path=data_path)
        except (TypeError, ApiException) as invalid_query_response:
            if (
                isinstance(invalid_query_response, ApiException)
                and invalid_query_response.status != 404
                and "bigquery"
                not in invalid_query_response.headers["Turbolinks-Location"]
            ):
                raise invalid_query_response
        else:
            metadata_file_path = f"{data_path}dataset-metadata.json"
            with open(metadata_file_path) as metadata_file:
                json_data = json.load(metadata_file)

            os.remove(metadata_file_path)

        return json_data

    def get_query_metadata(
        self,
        object_paths: Collection[str],
    ) -> pd.DataFrame:
        """Retrieves the metadata for the object_paths objects.

        Parameters
        ----------
        object_paths : str or list-like of str

        Returns
        -------
        metadata_df : pandas.DataFrame
        """

        object_paths = validate_metadata_parameters(object_paths)

        metadata_df = pd.DataFrame()

        for object_path in self.track_determinate_progress(object_paths):
            json_data = self._retrieve_object_json(object_path=object_path)
            metadata_df = pd.concat([metadata_df, pd.DataFrame(json_data)]).reset_index(
                drop=True
            )

        metadata_df = metadata_df.convert_dtypes()

        return metadata_df

    def get_all_metadata(self, search_dict: TermTypeResultDict) -> TermTypeResultDict:
        """Retrieves metadata for records contained in input DataFrames.

        Parameters
        ----------
        search_dict : dict of pandas.DataFrames
            Output from get_all_search_outputs function.

        Returns
        -------
        metadata_dict : dict of DataFrames
            Stores the results of each call to get_query_metadata in the
            form: metadata_dict[(search_term, search_type)] = df.
        """

        object_path_dict = {}

        for query, query_df in search_dict.items():
            # Only want to get metadata for non-empty dataset DataFrames
            if "kernels" not in query and query_df is not None:
                # Extract object paths
                object_paths = query_df.id.values
                object_path_dict[query] = object_paths

        metadata_dict = super()._get_metadata_from_paths(
            object_path_dict=object_path_dict
        )

        return metadata_dict



"""
Module for collecting data from Dryad repository.
"""

from collections.abc import Collection
from typing import Any, Optional, Union

import numpy as np
import pandas as pd

from ..base import (
    BaseCollector,
    BaseTermCollector,
)
from ..utils.validating import validate_metadata_parameters
from pycurator._typing import SearchTerm, TermResultDict


class DryadCollector(BaseTermCollector):
    """DataDryad collector for search term queries.

    Parameters
    ----------
    search_terms : list-like, optional
        Terms to search over. Can be (re)set via set_search_terms()
        or passed in directly to search functions.
    credentials : str, optional (default=None)
        JSON filepath containing credentials in form
        {repository_name}: {key}.
    """

    def __init__(
            self,
            search_terms: Optional[Collection[SearchTerm]] = None,
            credentials: Optional[bool] = None,
    ) -> None:

        super().__init__(
            repository_name="dryad", search_terms=search_terms, credentials=credentials
        )

        self.base_url = "https://datadryad.org/api/v2"
        self.merge_on = "version"

    @staticmethod
    def accepts_user_credentials() -> bool:
        return True

    @BaseCollector.track_indeterminate_progress
    def _conduct_search_over_pages(
            self,
            search_url: str,
            search_params: Any,
            print_progress: bool = False,
            delim: Optional[str] = None,
    ) -> pd.DataFrame:
        """Query records from the Dryad API for given parameters.

        Parameters
        ----------
        search_url : str
        search_params : dict
            Contains parameters to pass to requests.get({params}).
            Most common include search term 'q', and page index 'page'.
            For full details, see the Notes.
        print_progress : bool, optional (default=False)
            If True, updates on query page progress is sent to object
            queue to be displayed in UI window.
        delim : bool, optional (default=None)
            Key to grab results from query response JSON. If None,
            entire JSON return is considered as the data results.

        Returns
        -------
        search_df : pandas.DataFrame

        Notes
        -----
        Dryad allows the following parameters when querying the noted
        record type.

        Datasets:
        page : int, optional
            Page to search over.
        per_page : int, optional
            Number of results per page.
        q : str, optional
            Term to query for.
        affiliation : str, optional
            ROR identifier to require in a dataset's authors.
        tenant : str, optional
            Tenant organization in Dryad. Ignored if affiliation given.
        modifiedSince : str, optional
            An ISO 8601 UTC timestamp for limiting results.

        Files of a dataset version:
        id : int
            Version ID of the dataset.
        page : int, optional
            As above.
        per_page : int, optional
            As above.

        When searching for file metadata, only id, as described above,
        is allowed.
        """

        search_df = pd.DataFrame()

        if print_progress:
            self._update_query_ref(
                search_term=search_params["q"], page=search_params["page"]
            )
        _, output = self.get_request_output(url=search_url, params=search_params)

        while output.get("count"):
            output = output["_embedded"]

            if delim:
                output = output[delim]

            output_df = pd.DataFrame(output)
            output_df["page"] = search_params["page"]

            search_df = pd.concat([search_df, output_df]).reset_index(drop=True)

            search_params["page"] += 1
            if print_progress:
                self._update_query_ref(
                    search_term=search_params["q"], page=search_params["page"]
                )
            _, output = self.get_request_output(url=search_url, params=search_params)

        return search_df

    @BaseTermCollector.validate_search_term
    def get_individual_search_output(self, search_term: SearchTerm) -> pd.DataFrame:
        """Returns information about all datasets from DataDryad.

        Parameters
        ----------
        search_term : str

        Returns
        -------
        search_df : pandas.DataFrame

        Raises
        ------
        TypeError
            Incorrect search_term type.
        """

        search_url = f"{self.base_url}/search"
        search_params = {"q": search_term, "page": 1, "per_page": 100}

        search_df = self._conduct_search_over_pages(
            search_url=search_url,
            search_params=search_params,
            print_progress=True,
            delim="stash:datasets",
        )

        # Add dataset-specific version id for metadata querying
        search_df["version"] = self._extract_version_ids(search_df)

        return search_df

    def get_query_metadata(
            self, object_paths: Union[str, Collection[str]], **kwargs: Any
    ) -> pd.DataFrame:
        """Retrieves the metadata for the object_paths objects.

        Parameters
        ----------
        object_paths : str or collection of str

        Returns
        -------
        metadata_df : pandas.DataFrame

        Raises
        ------
        TypeError
            If no object paths are provided.
        """

        object_paths = validate_metadata_parameters(object_paths)

        start_page = 1
        metadata_df = pd.DataFrame()

        for object_path in self.track_determinate_progress(object_paths):
            search_url = f"{self.base_url}/versions/{object_path}/files"
            search_params = {"page": start_page}

            object_df = self._conduct_search_over_pages(
                search_url=search_url,
                search_params=search_params,
                delim="stash:files",
                print_progress=False,
            )

            object_df["version"] = object_path
            object_df.loc[:, "page"] = search_params["page"]
            metadata_df = pd.concat([metadata_df, object_df]).reset_index(drop=True)

        return metadata_df

    @staticmethod
    def _extract_version_ids(version_df: pd.DataFrame) -> pd.DataFrame:
        """Retrieve ids from DataFrame entries."""
        return version_df["_links"].apply(
            lambda entry: entry.get("stash:version", {}).get("href", "").split("/")[-1]
            if entry is not np.nan
            else None
        )

    def get_all_metadata(self, search_dict: TermResultDict) -> TermResultDict:
        """Retrieves metadata for records contained in input DataFrames.

        Parameters
        ----------
        search_dict : dict
            Dictionary of DataFrames from get_all_search_outputs.

        Returns
        -------
        metadata_dict : dict
        """

        object_path_dict = {
            query: self._extract_version_ids(version_df=query_df)
            for query, query_df in search_dict.items()
        }

        return self._get_metadata_from_paths(object_path_dict)



"""
Module for UI page base class.
"""

from abc import abstractmethod
from tkinter import ttk
from typing import Any, Callable, NoReturn


class ViewPage(ttk.Frame):
    """Base for page in UI View object.

    Parameters
    ----------
    *args : tuple, optional
        Positional arguments to pass to ttk.Frame constructor.
    **kwargs : dict, optional
        Keyword arguments to pass to ttk.Frame constructor.

    Attributes
    ----------
    controller : MVC.CuratorController
        Default of None to allow UI instantiation of pages before
        Controller.
    is_initialized : bool
    next_page_button : None or ttk.Button
        Widget with action linking to the next sequential page of the UI.

    Methods
    -------
    set_controller
    no_overwrite
        Decorator for loading a page without overwriting the previously
        displayed values. Useful for reloading a page without destroying
        widgets or reconstructing an instance of the page.
    reset_frame
    attempt_raise
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.next_page = kwargs.pop("next_page", None)
        ttk.Frame.__init__(self, *args, **kwargs)

        self.controller = None
        self.is_initialized = False
        self.next_page_button = None

    def set_controller(self, controller: Any) -> None:
        """Setter for Controller element of PyCurator UI."""
        self.controller = controller

    @staticmethod
    def no_overwrite(show_func: Callable[[Any], Any]) -> Callable[[Any], Any]:
        """Wrapper for page display without element overwrite."""

        def display(self) -> None:
            raised = self.attempt_raise()
            if not raised:
                show_func(self)
                self.is_initialized = True

        return display

    @abstractmethod
    def reset_frame(self) -> NoReturn:
        """Abstract placeholder method for resetting page frame."""
        raise NotImplementedError

    def attempt_raise(self) -> bool:
        """Display page if initialized."""
        if self.is_initialized:
            try:
                self.reset_frame()
            except AttributeError:
                pass

            self.tkraise()
            return True
        return False



"""
Import collector format modules and create valid collector dict.
"""

from . import base, term_collectors, term_type_collectors, type_collectors

available_repos = {
    "Dryad": term_collectors.DryadCollector,
    "Zenodo": term_collectors.ZenodoCollector,
    "Dataverse": term_type_collectors.DataverseCollector,
    "Figshare": term_type_collectors.FigshareCollector,
    "Papers With Code": term_type_collectors.PapersWithCodeCollector,
}

try:
    from .term_type_collectors import KaggleCollector

    available_repos["Kaggle"] = KaggleCollector
except ImportError:
    pass

try:
    from .type_collectors import OpenMLCollector

    available_repos["OpenML"] = OpenMLCollector
except ImportError:
    pass

available_repos = dict(map(tuple, sorted(available_repos.items())))



"""
Module for utility functions for UI.
"""

from __future__ import annotations

import tkinter as tk
from tkinter import filedialog as fd
from tkinter import ttk
from typing import ParamSpec, Type, TypeVar, TYPE_CHECKING

if TYPE_CHECKING:
    from . import view

P = ParamSpec("P")
T = TypeVar("T")


def widget_label_frame(
    frame_master: tk.Misc,
    label_text: float | str,
    widget_cls: Type[tk.Widget],
    **widget_kwargs: P.kwargs,
) -> ttk.Frame:
    """Create frame containing vertically-aligned widget and label.

    Parameters
    ----------
    frame_master : tk.Misc
    label_text : float or str
    widget_cls : tk.Widget
    **widget_kwargs

    Returns
    -------
    _frame : ttk.Frame

    Examples
    --------
    >>> root = tk.Tk()
    >>> widget_label_frame(
        frame_master=root,
        label_text="Don't click this button:",
        widget_cls=ttk.Button,
        widget_text='Click me!',
        widget_command=sys.exit
    )
    """

    _frame = ttk.Frame(master=frame_master)
    _label = ttk.Label(master=_frame, text=label_text)
    _widget = widget_cls(master=_frame, **widget_kwargs)

    _label.grid(row=0, column=0)
    _widget.grid(row=0, column=1)
    return _frame


def select_from_files(
    root: view.page.ViewPage,
    selection_type: str,
    filetypes: list[tuple[str, str]] = (("All File Types", "*.*"),),
) -> None:
    """Allows user to select local file/directory.

    Parameters
    ----------
    root : Tkinter.Frame
        Derived Frame class that contains UI.
        If choosing file, frame_master must contain "files" dict attribute to
        hold the selected file.
    selection_type : str
        Type of selection to be made.
        Examples include "credentials", "directory", "css_paths" etc.
    filetypes : list of two-tuples of str,
            optional (default = [('All File Types', '*.*')])
        Allows users to input filetype restrictions.
        Each tuple should be of the form:
            ('File Type', '{file_name}.{file_format}')

    Raises
    ------
    NotImplementedError
        Occurs when called on a frame_master that does not contain a files
        attribute.

    Examples
    --------
    >>> f = tk.Frame()
    >>> select_from_files(
    ...     root=f,
    ...     selection_type='dir'
    ... )
    Traceback (most recent call last):
        ...
    NotImplementedError: 'root' must include add_run_parameter() method.'

    >>> from pycurator.gui import view
    >>> view_page = view.page.ViewPage()
    >>> select_from_files(
    ...     root=view_page,
    ...     selection_type='dir',
    ... )
    """

    if "dir" in selection_type:
        selection = fd.askdirectory(title="Choose Directory", initialdir=".")
    else:
        selection = fd.askopenfilename(
            title="Choose File", initialdir=".", filetypes=filetypes
        )

    try:
        root.controller.add_run_parameter(selection_type, selection)
    except AttributeError as no_run_param_method:
        raise NotImplementedError(
            "'root' must include add_run_parameter() method."
        ) from no_run_param_method



"""
CLI for running PyCurator user interface.
"""

from pycurator.gui import interface


def main():
    """Launch Pycurator UI for collecting data from repositories."""
    root = interface.PyCuratorUI()
    root.mainloop()


if __name__ == "__main__":
    main()



"""Module for creating View of MVC structure."""

from __future__ import annotations

import itertools as it
import tkinter as tk
from tkinter import ttk
from typing import Type, TYPE_CHECKING

from . import landing_page, run_page, selection_page, page

if TYPE_CHECKING:
    from .. import controller


class CuratorView(ttk.Frame):
    """View for the PyCurator UI.

    Parameters
    ----------
    parent : tk.Tk

    Attributes
    ----------
    pages : dict[str, ViewPage]
        Map from page name to page object.
    current_page : ViewPage

    See Also
    --------
    tk.Frame
    tk.Tk
    """

    def __init__(self, parent: tk.Tk) -> None:
        super().__init__(parent)
        self.current_page = None

        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)
        self.grid(row=0, column=0, sticky="nsew")

        page_order = (
            landing_page.LandingPage,
            selection_page.SelectionPage,
            run_page.RunPage,
        )
        self.pages = {
            C: C(self, next_page=N) for C, N in it.pairwise(page_order + (None,))
        }

    def set_controller(self, curator_controller: controller.CuratorController) -> None:
        """Setter for UI Controller of View object."""
        for page_ in self.pages.values():
            page_.set_controller(curator_controller)

    def show(self, page_: Type[page.ViewPage] = landing_page.LandingPage) -> None:
        """Display the provided page on the UI."""
        self.current_page = self.pages[page_]
        self.current_page.show()



"""
Module for collecting data from Figshare repository.
"""

from collections.abc import Collection
from typing import Optional, Union

import pandas as pd

from ..base import BaseCollector, BaseTermTypeCollector
from ..utils.validating import validate_metadata_parameters
from pycurator._typing import SearchTerm, SearchType, TermTypeResultDict


class FigshareCollector(BaseTermTypeCollector):
    """Figshare collector for search term and type queries.

    Parameters
    ----------
    search_terms : list-like, optional (default=None)
        Terms to search over. Can be (re)set via set_search_terms() or
        passed in directly to search functions.
    search_types : list-like, optional (default=None)
        Data types to search over. Can be (re)set via set_search_types()
        or passed in directly to search functions to override set
        parameter.
    credentials : str, optional (default=None)
        JSON filepath containing credentials in form
        {repository_name}: {key}.
    """

    def __init__(
            self,
            search_terms: Optional[Collection[SearchTerm]] = None,
            search_types: Optional[Collection[SearchType]] = None,
            credentials: Optional[str] = None,
    ) -> None:
        super().__init__(
            repository_name="figshare",
            search_terms=search_terms,
            search_types=search_types,
            credentials=None,
        )
        self.base_url = "https://api.figshare.com/v2"
        self.merge_on = "id"
        self.headers = {}

        if credentials:
            self.credentials = self.load_credentials(credential_filepath=credentials)

    @staticmethod
    def accepts_user_credentials() -> bool:
        return True

    @classmethod
    @property
    def search_type_options(cls) -> tuple[SearchType, ...]:
        return ("articles", "collections", "projects")

    def load_credentials(self, credential_filepath: str) -> Union[str, None]:
        """Load the credentials given filepath or token.

        Parameters
        ----------
        credential_filepath : str
            JSON filepath containing credentials in the form
            {repository_name}: {key}.

        Returns
        -------
        credentials : str or None
        """

        credentials = super().load_credentials(credential_filepath=credential_filepath)
        self.headers["Authorization"] = f"token {credentials}"
        return credentials

    @BaseCollector.track_indeterminate_progress
    @BaseTermTypeCollector.validate_term_and_type
    def get_individual_search_output(
            self,
            search_term: SearchTerm,
            search_type: SearchType,
    ) -> Union[TermTypeResultDict, None]:
        """Queries the Figshare API for the given search term and type.

        Parameters
        ----------
        search_term : str
        search_type : {'articles', 'collections', 'projects'}

        Returns
        -------
        search_df : pandas.DataFrame

        Raises
        ------
        TypeError
            Incorrect search_term type.
        ValueError
            Invalid search_type provided.
        """

        start_page = 1
        page_size = 1000
        search_year = 1950
        search_date = f"{search_year}-01-01"
        search_url = f"{self.base_url}/{search_type}"

        search_df = pd.DataFrame()
        output_df = None

        search_params = {
            "search_for": search_term,
            "published_since": search_date,
            "order_direction": "asc",
            "page": start_page,
            "page_size": page_size,
        }

        response, output = self.get_request_output_and_update_query_ref(
            url=search_url,
            params=search_params,
            headers=self.headers,
            published_since=search_date,
            page=start_page,
        )

        while response.status_code == 200:
            while response.status_code == 200 and output:
                output_df = pd.DataFrame(output)
                output_df["search_page"] = search_params["page"]
                output_df["publish_query"] = search_params["published_since"]

                search_df = pd.concat([search_df, output_df]).reset_index(drop=True)

                search_params["page"] += 1
                response, output = self.get_request_output_and_update_query_ref(
                    url=search_url,
                    params=search_params,
                    headers=self.headers,
                    published_since=search_params["published_since"],
                    page=search_params["page"],
                )

            if (
                    output_df is not None
                    and output_df.shape[0] < search_params["page_size"]
            ):
                return search_df

            # Get new date to search
            search_date = search_df["published_date"].values[-1].split("T")[0]
            search_params["published_since"] = search_date
            search_params["page"] = start_page

            # Conduct search
            response, output = self.get_request_output_and_update_query_ref(
                url=search_url,
                params=search_params,
                headers=self.headers,
                published_since=search_params["published_since"],
                page=search_params["page"],
            )

        return search_df

    def get_query_metadata(
            self,
            object_paths: Union[str, Collection[str]],
    ) -> pd.DataFrame:
        """
        Retrieves metadata for the object_paths objects.

        Parameters
        ----------
        object_paths : str or list-like
            string or list of strings containing object paths.

        Returns
        -------
        metadata_df : pandas.DataFrame
            Metadata for the requested objects.
        """

        object_paths = validate_metadata_parameters(object_paths)

        metadata_df = pd.DataFrame()

        for object_path in self.track_determinate_progress(object_paths):
            _, json_data = self.get_request_output(
                url=object_path, headers=self.headers
            )

            metadata_df = pd.concat(
                [metadata_df, pd.DataFrame([json_data])]
            ).reset_index(drop=True)

        return metadata_df

    def get_all_metadata(self, search_dict: TermTypeResultDict) -> TermTypeResultDict:
        """Retrieves metadata for records contained in input DataFrames.

        Parameters
        ----------
        search_dict : dict
            Dictionary of DataFrames from get_all_search_outputs.

        Returns
        -------
        metadata_dict : dict of tuple of str to dataframe
        """

        object_path_dict = {}

        for query, query_df in search_dict.items():
            if query_df is not None:
                _, search_type = query
                object_ids = query_df.id.convert_dtypes().tolist()
                object_paths = [
                    f"{self.base_url}/{search_type}/{object_id}"
                    for object_id in object_ids
                ]

                object_path_dict[query] = object_paths

        return self._get_metadata_from_paths(object_path_dict)
